# -*- coding: utf-8 -*-
"""
Created on Thu Jul  4 16:25:54 2019

@author: Pavel Gostev
"""
import gc
from dataclasses import dataclass

import lecroyparser
from . import tekwfm as tek

from scipy.sparse.linalg import spsolve
from scipy import sparse
import numpy as np
from os.path import isfile, join
from os import listdir
import time
from compress_pickle import dump, load
from joblib import Parallel, delayed

gc.enable()

VENDOR_EXT = {'tek': 'wfm',
              'lecroy': 'trc'}


def parse_file(datafile, vendor):
    if vendor == 'lecroy':
        data = lecroyparser.ScopeData(datafile)
        delattr(data, "file")
    if vendor == 'tek':
        data = tek.ScopeData(datafile)
    return data


def list_files(datadir, vendor, fsoffset, fsnum):
    lf = [join(datadir, f) for f in listdir(datadir)
          if isfile(join(datadir, f)) and f.endswith(VENDOR_EXT[vendor])]
    if fsoffset != 0:
        lf = lf[fsoffset:]
    if fsnum > 0:
        lf = lf[:fsnum]
    return lf


def parse_files(oscfiles, vendor, fsnum=0, parallel=False):

    if fsnum > 0:
        oscfiles = oscfiles[:fsnum]
    if parallel:
        data = Parallel(
            n_jobs=-1)([delayed(parse_file)(df, vendor) for df in oscfiles])
    else:
        data = [parse_file(df, vendor) for df in oscfiles]
    return data


def windowed(data, div_start, div_width):
    div_points = len(data.x) / data.horizInterval
    wstart = int(div_start * div_points)
    wwidth = int(div_width * div_points)
    return data.x[wstart:wstart + wwidth], data.y[wstart:wstart + wwidth]


def baseline_als(y, lam, p, niter=2):
    """
    See "Asymmetric Least Squares Smoothing" by P. Eilers and H. Boelens in 2005
    https://www.researchgate.net/publication/228961729_Baseline_Correction_with_Asymmetric_Least_Squares_Smoothing
    """
    L = len(y)
    D = sparse.diags([1, -2, 1], [0, -1, -2], shape=(L, L - 2))
    w = np.ones(L)
    for i in range(niter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.T)
        z = spsolve(Z, w * y)
        w = p * (y > z) + (1 - p) * (y < z)
    return z


def correct_baseline(y, lam=1e5, p=0.1):
    baseline_values = baseline_als(y, lam=lam, p=p)
    return y - baseline_values


def single_pulse(y):
    """
    Get amplitude of the single pulse

    Parameters
    ----------
    y : ndarray
        Oscillogram of the single pulse.

    Returns
    -------
    Amplitude : float
        Maximum amplitude of the pulse.

    """
    return max(y)


def periodic_pulse(data, frequency, time_window):
    discretedata = []
    points_period = int(1 / frequency / data.horizInterval) + 1
    points_window = int(time_window / data.horizInterval) + 1
    y = data.y
    init_point = np.argmax(y)
    pulses_points = np.append(
        np.arange(init_point, 0, -points_period)[::-1],
        np.arange(init_point, len(y), points_period))
    for p in pulses_points:
        if p < points_window:
            low = 0
            top = points_window
        else:
            low = p - points_window // 2
            top = p + points_window // 2
        discretedata.append(single_pulse(y[low:top]))
    return discretedata


def scope_unwindowed(data, time_discrete):
    points_discrete = int(time_discrete // data.horizInterval)
    y = data.y
    points_discrete += 1
    discretedata = [single_pulse(y[i:i + points_discrete])
                    for i in range(0, len(y), points_discrete)]
    return discretedata


def memo_oscillogram(data, vendor, correct_bs=True):
    if type(data) is str:
        filedata = (data, parse_file(data, vendor))
    if type(data) == tuple:
        if type(data[1]) is not str:
            return data

        filedata = (data[0], parse_file(data[0], vendor))
        filedata[1].y = data[1]
        return filedata

    y = filedata[1].y
    y -= np.min(y)
    if correct_bs:
        filedata[1].y = correct_baseline(y)
    else:
        filedata[1].y = y
    return filedata


@dataclass
class PulsesHistMaker:
    """
    Class for extract amplitude histogram from plenty of oscillogram files
    Supported formats are:
        - LecCroy binary .trc
        - Tektronix binary v.2 .wfm

    __init__ arguments
    ------------------

    datadir: str
        Path to oscillogram files
    fsnum: int = -1
        Number of files to process.
        If -1 all files in the datadir will be processed
    fsoffset: int = 0
        Skip files from the first ine in the datadir
    fchunksize: int = 10
        The size of chunk for parallel peocessing
    parallel: bool = False
        Use joblib or not?
    parallel_jobs: int = -1
        Number of parallel jobs
    memo_file: str = ''
        Path to file with compressed data from previous processing.
        Can be use iff use save it previously with 'save_memo' method
    histbins: int = 2000
        Numbere of bins in the histogram.
        Can be change in methods 'make_hist' and 'get_hist'
    correct_baseline: bool = True
        Correct baseline of not?
    vendor: str = 'lecroy'
        Vendor name. Can be 'tek' or 'lecroy'
    """

    datadir: str
    fsnum: int = -1
    fsoffset: int = 0
    fchunksize: int = 10
    parallel: bool = False
    parallel_jobs: int = -1
    memo_file: str = ''
    histbins: int = 2000
    correct_baseline: bool = True

    vendor: str = 'lecroy'
    vendors: tuple = ('lecroy', 'tek')

    def __post_init__(self):
        if not self.parallel:
            self.parallel_jobs = 1
        if self.vendor not in self.vendors:
            raise ValueError('vendor must be in %s, not %s' %
                             (self.vendors, self.vendor))

    def read(self, fsnum=-1, parallel_read=False):
        if fsnum == -1:
            fsnum = self.fsnum
        self.rawdata = list_files(
            self.datadir, self.vendor, self.fsoffset, fsnum)
        if self.memo_file:
            with open(self.memo_file, 'rb') as f:
                memodata = load(f, compression='lzma',
                                set_default_extension=False)
            for k, path in enumerate(self.rawdata):
                if path in self.rawdata:
                    self.rawdata[k] = (path, memodata[path])

        self.filesnum = len(self.rawdata)

    def save_memo(self, filename):
        self.clear_rawdata()
        with open(filename, 'wb') as f:
            dump(dict(self.rawdata), f, compression='lzma',
                 set_default_extension=False)

    def save_hist(self, filename):
        np.savetxt(filename, np.vstack((self.bins[:-1], self.hist)).T)

    def clear_rawdata(self):
        for k in range(len(self.rawdata)):
            p, d = self.rawdata[k]
            self.rawdata[k] = (p, d.y)

    def single_pulse_hist(self, div_start=5.9, div_width=0.27):
        self.discretedata = []
        i = 1
        for d in self.rawdata:
            x, y = windowed(d, div_start, div_width)
            self.discretedata.append(single_pulse(y))
            i += 1
        self.make_hist(self.histbins)

    def multi_pulse_histogram(self, frequency=2.5e6, time_window=7.5e-9):
        self.parse(periodic_pulse, (frequency, time_window))
        self.make_hist(self.histbins)

    def unwindowed_histogram(self, time_discrete=15e-9):
        self.parse(scope_unwindowed, (time_discrete,))
        self.make_hist(self.histbins)

    def parse(self, func, args):
        self.discretedata = []

        for i in range(0, self.filesnum, self.fchunksize):
            t = time.time()
            hb = min(i + self.fchunksize, self.filesnum)
            self.rawdata[i:hb] = Parallel(n_jobs=self.parallel_jobs)([
                delayed(memo_oscillogram)(
                    df, self.vendor, self.correct_baseline)
                for df in self.rawdata[i:hb]])
            pulsesdata = [func(df[1], *args) for df in self.rawdata[i:hb]]
            self.discretedata += pulsesdata

            print('Files ##%d-%d time %.2f s' %
                  (i, hb, time.time() - t), end='\t')

            del pulsesdata
            gc.collect()

    def make_hist(self, histbins):
        self.hist, self.bins = np.histogram(self.discretedata, bins=histbins)

    def get_hist(self, histbins):
        self.make_hist(histbins)
        return self.bins[:-1], self.hist



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 20:51:49 2019

@author: Pavel Gostev
"""
import numpy as np

from scipy.optimize import brute, OptimizeResult
from . import (g2, mean, normalize, P2Q, moment, fact)
import logging


log = logging.getLogger('sc_crosstalk')
log.setLevel(logging.INFO)
info = log.info


def ppoisson(pmean, N):
    return np.array([pmean ** n * np.exp(-pmean) / fact(n) for n in range(N)])


def d_crosstalk_4n(p_crosstalk):
    """
    Probabilities of k ≤ 5 triggered pixels for the 4-neighbours model

    See table 1 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    ctnoise : ndarray
        Probability distributions of the total number
        of triggered pixels for a single initially fired pixel.
    """

    q_crosstalk = (1 - p_crosstalk)
    ctnoise = np.zeros(6)
    ctnoise[1] = q_crosstalk ** 4
    ctnoise[2] = 4 * p_crosstalk * q_crosstalk ** 6
    ctnoise[3] = 18 * p_crosstalk ** 2 * q_crosstalk ** 8
    ctnoise[4] = 4 * p_crosstalk ** 3 * q_crosstalk ** 8 * (
        1 + 3 * q_crosstalk + 18 * q_crosstalk ** 2)
    ctnoise[5] = 5 * p_crosstalk ** 4 * q_crosstalk ** 10 * (
        8 + 24 * q_crosstalk + 55 * q_crosstalk ** 2)
    return ctnoise


def p_crosstalk_m(m, k, p_crosstalk):
    """
    The probability of total number k of triggered pixels provided m primaries

    See formula 2.20 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    m : int
        Number of primary pixels.
    k : int
        Number of triggered pixels.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    P_{m+1} : float
        The probability of total number k of triggered pixels
        provided m primaries

    """

    ctnoise = d_crosstalk_4n(p_crosstalk)
    if m == 1:
        if k > 5:
            return 0
        return ctnoise[k]
    return sum(p_crosstalk_m(m - 1, k - i, p_crosstalk) * ctnoise[i]
               for i in range(1, k - m + 2, 1) if i <= 5)


def p_crosstalk_m_pyomo(m, k, distr_crosstalk):
    """
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    Model with 4 neighbors with saturation
    See formula (14)
    """
    if m == 1:
        if k > 4:
            return 0
        return distr_crosstalk[k]
    return sum(p_crosstalk_m_pyomo(m - 1, k - i - 1, distr_crosstalk) * distr_crosstalk[i]
               for i in range(1, k - m + 2, 1) if i <= 4)


def distort(Qcorr, p_crosstalk):
    """
    Include crosstalk noise into photocounting statistics.
    We use model with 4 neighbors with saturation.

    See formula 2.19 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    Qcorr : iterable
        The photocounting statistics.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    Q : ndarray
        Noised photocounting statistics.

    """

    N = len(Qcorr)
    Q = np.zeros(N)
    Q[0] = Qcorr[0]

    def genf(k):
        return sum(Qcorr[m] * p_crosstalk_m(m, k, p_crosstalk) for m in range(1, k + 1, 1))
    Q[1:] = [genf(k) for k in range(1, N, 1)]
    return Q

def compensate(Q, p_crosstalk):
    """
    Remove crosstalk noise from photocounting statistics.
    We use model with 4 neighbors with saturation.

    See formula 2.21 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    Q : iterable
        The photocounting statistics.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    Qcorr : ndarray
        Denoised photocounting statistics.

    """
    N = len(Q)
    eps = total_pcrosstalk(p_crosstalk)
    Qcorr = np.zeros(N)
    Qcorr[0] = Q[0]
    Qcorr[1] = Q[1] / (1 - eps)
    for m in range(2, N, 1):
        Qcorr[m] = (1 - eps) ** -m * \
            (Q[m] - sum(Qcorr[k] * p_crosstalk_m(k, m, p_crosstalk)
                        for k in range(1, m, 1)))
    return normalize(np.abs(Qcorr))


def optctp(_pct_param, Q, PDE, N, mtype, n_cells):
    p_crosstalk, poisson_mean = _pct_param
    pm = P2Q(ppoisson(poisson_mean, N), PDE, len(Q),
             mtype=mtype, n_cells=n_cells)
    est = distort(pm, p_crosstalk)
    gdelta = (g2(est) - g2(Q)) ** 2
    return gdelta


def optimize_pcrosstalk(Q, PDE, N, mtype='binomial', n_cells=0, Ns=100):
    """
    Brute searching of crosstalk probability
    by optimizing of g2 difference from noised data and
    noised poisson photocounting statistics.
    We use model with 4 neighbors with saturation.

    See
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    Q : iterable
        Experimental photocounting statistics of a laser source.
    PDE : float
        PDE of the detector.
    N : int
        Size of poisson photon-number statistics.
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in the most of applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells in the subbinomial case. The default is 0.
    Ns : int, optional
        Size of calculation grid of brute function.
        The default is 100.

    Returns
    -------
    p_ct: float
        The probability of a single crosstalk event.

    res: OptimizeResult
        See scipy.optimize.brute description.
        res.x0 consists of optimal p_crosstalk and
        optimal poisson photon-number distribution
        p_crosstalk is in res.x0[0]

    """

    res = brute(optctp, ([0.00, 0.01], [mean(Q) / PDE * 0.9,
                                        mean(Q) / PDE * 1.1]),
                args=(Q, PDE, N, mtype, n_cells), Ns=Ns, full_output=True)
    info("P_ct = {r[0][0]}, Δg(2) = {r[1]}".format(r=res))
    return res[0][0], OptimizeResult(x=res[0], fval=res[1],
                                     grid=res[2], Jout=res[3])


def Q2total_pcrosstalk(Q):
    """
    Calculate model independent total crosstalk probability.
    It's may be vary from the result of optimize_pcrosstalk

    See formula 2.23 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    Q : iterable
        Experimental photocounting statistics of a pulsed laser source.

    Returns
    -------
    epsilon : float
        The total crosstalk probability.

    """

    mu = - np.log(Q[0])
    return 1 - Q[1] / mu / np.exp(- mu)


def total_pcrosstalk(p_crosstalk):
    """
    Calculate total crosstalk probability from
    the probability of single crosstalk event.
    We use model with 4 neighbors with saturation

    See text after formula 2.1 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    epsilon : float
        The total crosstalk probability.

    """

    return 1 - (1 - p_crosstalk)**4


def ENF(p_crosstalk):
    """
    Calculate excess noise factor (ENF) of the detector

    See formulas 2.16, 2.17 and 2.27 in
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    ENF : float
        Excess noise factor.

    """

    d = d_crosstalk_4n(p_crosstalk)
    r = d[5] / (1 - sum(d[1:5]))
    e1 = moment(d[:5], 1) + d[5] * (1 + 4*r) / r ** 2
    var1 = moment(d[:5], 2) + d[5] * (2 + 7*r + 16 * r ** 2) / r ** 3 - e1 ** 2
    return 1 + var1 / e1 ** 2



# -*- coding: utf-8 -*-
"""
@author: Pavel Gostev
"""
from ._numpy_core import normalize, lrange, abssum

import numpy as np
from scipy.signal import find_peaks
from scipy.special import eval_hermitenorm
from scipy.optimize import minimize
import matplotlib.pyplot as plt


def loadhist(path, **kwargs):
    for dl in [',', ' ', '\t']:
        try:
            bins, hist = np.loadtxt(path, delimiter=dl, unpack=True, **kwargs)
        except (ValueError, TypeError):
            continue
        finally:
            return bins, hist
    raise ValueError("Can't parse the histogram from the file %s" % path)


def gauss_hermite_poly(x, norm_factor, peak_pos, sigma, h3, h4):
    w = (x - peak_pos) / sigma
    return norm_factor * np.exp(- w ** 2 / 2)  * (1 + h3*eval_hermitenorm(3, w) + h4*eval_hermitenorm(4, w))


def peak_area(norm_factor, peak_pos, sigma, h3, h4):
    return norm_factor * sigma * (np.sqrt(2*np.pi) + h4)


def minpoly(popt, bins, hist):
    return abssum(gauss_hermite_poly(bins, *popt) - hist)


def mindowns(popt, bins, hist, downs):
    return abssum((hist - gauss_hermite_poly(bins, *popt))[downs])


def hist2Q(hist, bins, discrete, 
           threshold=1, peak_width=1, 
           plot=False, method='sum'):
    """
    Build photocounting statistics from an experimental histogram
    by gaussian-hermite polynoms or simple sum

    Parameters
    ----------
    hist : iterable
        The experimental histogram.
    discrete : int
        The amplitude of single photocount pulse in points.
    threshold : int, optional
        Minimal number of events to find histogram peak.
        The default is 1.
    peak_width : int, optional
        The width of peaks.
        It must be 1 if the histogram is made by 'count' method.
        It must be greater if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    plot : bool, optional
        Flag to plot hist and results of find_peaks.
        The default is False.
    method: {'sum', 'fit'}
        Method of the photocounting statistics construction.
            'sum' is a simple summation between minimums of the histogram

            'fit' is a gauss-hermite function fitteing like in [1]

    Returns
    -------
    Q : ndarray
        The photocounting statistics.

    References
    ----------
    .. [1]
    Ramilli, Marco, et al. "Photon-number statistics with silicon photomultipliers."
    JOSA B 27.5 (2010): 852-862.

    """
    downs, _ = find_peaks(-hist, distance=discrete, width=1)
    if plot:
        plt.plot(bins, hist)
        
    dsum = 0
    while sum(hist[downs]) > 0.01:

        dres = minimize(mindowns, args=(bins, hist, downs),
                        x0=(1, bins[len(bins) // 2], np.sqrt(bins[len(bins) // 2]), 0, 0))
        hist -= gauss_hermite_poly(bins, *dres.x)
        hist[hist < 0] = 0
        if dsum <= sum(hist[downs]):
            break
        dsum = sum(hist[downs])
        
    if plot:
        plt.plot(bins, hist)

    peaks, _ = find_peaks(hist, threshold=threshold, distance=discrete,
                          width=peak_width)
    downs, _ = find_peaks(-hist, distance=discrete, width=1)
    downs = np.sort(np.append([downs[0] - int(discrete * 1.2), len(bins) - 1], downs))
    if plot:
        plt.scatter(bins[peaks], hist[peaks])
        plt.scatter(bins[downs], hist[downs])
    plt.show()

    Q = []
    for i in lrange(downs)[:-1]:
        for p in peaks:
            dl = downs[i]
            dt = downs[i+1]
            if p > dl and p < dt:
                if method == 'sum':
                    Q.append(sum(hist[dl:dt]))
                if method == 'fit':
                    res = minimize(minpoly, args=(bins[dl:dt], hist[dl:dt]), tol=1e-16,
                                   x0=(1, bins[p], bins[1] - bins[0], 0.01, 0.01),
                                   bounds=list(zip([1, bins[dl], bins[1] - bins[0], -0.01, -0.01], 
                                                   [hist[p], bins[dt], np.sqrt(bins[dt] - bins[dl]) / 4, 0.01, 0.01])))
                    popt = res.x
                    Q.append(peak_area(*popt))
                    if plot:
                        plt.plot(bins, gauss_hermite_poly(bins, *popt))
    if plot and method == 'fit': 
        #plt.plot(bins, hist)
        plt.show()
    return normalize(Q)


class QStatisticsMaker:
    """
    Class to make photocounting statistics from histogram

    __init__ arguments
    ----------
    fname : string
        File name contains the histogram.
    photon_discrete : float
        The amplitude of the single-photocount pulse.
    peak_width : int, optional
        The width of peaks.
        It must be 1 if the histogram is made by 'count' method.
        It must be greater if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    method: {'sum', 'fit'}
        Method of the photocounting statistics construction.
            'sum' is a simple summation between minimums of the histogram
            
            'fit' is a gauss-hermite function fitteing like in [1]
    skiprows : int, optional
        Number of preamble rows in the file. The default is 0.
    plot : bool, optional
        Flag to plot hist and results of find_peaks.
        The default is False.
        
    References
    ----------
    .. [1]
    Ramilli, Marco, et al. "Photon-number statistics with silicon photomultipliers."
    JOSA B 27.5 (2010): 852-862.

    """

    def __init__(self, fname, photon_discrete,
                 peak_width=1, method='fit', skiprows=0, plot=False):
        self.photon_discrete = photon_discrete
        self.fname = fname
        self.plot = plot

        self._extract_data(skiprows)
        self.Q = hist2Q(self.hist, self.bins, discrete=self.points_discrete // 1.2,
                        peak_width=peak_width, plot=self.plot, method=method)

    # Reading information from the file
    def _extract_data(self, skiprows):
        self.bins, self.hist = loadhist(self.fname, skiprows=skiprows)
        if self.bins[1] - self.bins[0] < 0:
            self.bins = self.bins[::-1]
            self.hist = self.hist[::-1]
        self.points_discrete = int(self.photon_discrete /
                                   (self.bins[1] - self.bins[0]))

    def getq(self):
        """
        Returns the photocounting statistics was made

        Returns
        -------
        self.Q : ndarray
            self.Q[self.Q > 0].

        """

        self.Q = self.Q[self.Q > 0]
        return normalize(self.Q)



# wfm reader proof-of-concept
# https://www.tek.com/sample-license
# reads volts vs. time records (including fastframes) from little-endian version 2 WFM files

# See Also
# Performance Oscilloscope Reference Waveform File Format
# https://download.tek.com/manual/Waveform-File-Format-Manual-077022011.pdf

# See also
# https://forum.tek.com/viewtopic.php?t=141081

import struct
import numpy as np


class WfmReadError(Exception):
    """error for unexpected things"""
    pass


def read_wfm(target):
    """return sample data from target WFM file"""
    with open(target, 'rb') as f:
        hbytes = f.read(838)
        meta = decode_header(hbytes)
        # file signature checks
        if meta['byte_order'] != 0x0f0f:
            raise WfmReadError('big-endian not supported in this version')
        if meta['version'] != b':WFM#002':
            raise WfmReadError('only version 2 wfms supported in this version')
        if meta['imp_dim_count'] != 1:
            raise WfmReadError('imp dim count not 1')
        if meta['exp_dim_count'] != 1:
            raise WfmReadError('exp dim count not 1')
        if meta['record_type'] != 2:
            raise WfmReadError('not WFMDATA_VECTOR')
        # if meta['exp_dim_1_type'] != 0:
        #    raise WfmReadError('not EXPLICIT_SAMPLE')
        if meta['time_base_1'] != 0:
            raise WfmReadError('not BASE_TIME')
        if meta['fastframe']:
            raise WfmReadError('Fast Frames are not supported')
        # read curve block
        bin_wave = np.memmap(filename=f,
                             dtype=meta['dformat'],
                             mode='r',
                             offset=meta['curve_offset'],
                             shape=(meta['dlen']),
                             order='F')
        # close file
    # slice out buffer values
    scaled_array = bin_wave * meta['vscale'] + meta['voffset']
    return scaled_array, meta['tstart'], meta['tscale'], meta['tfrac']


def decode_header(header_bytes):
    """returns a dict of wfm metadata"""
    wfm_info = {}
    if len(header_bytes) != 838:
        raise WfmReadError('wfm header bytes not 838')
    wfm_info['byte_order'] = struct.unpack_from('H', header_bytes, offset=0)[0]
    wfm_info['version'] = struct.unpack_from('8s', header_bytes, offset=2)[0]
    wfm_info['imp_dim_count'] = struct.unpack_from(
        'I', header_bytes, offset=114)[0]
    wfm_info['exp_dim_count'] = struct.unpack_from(
        'I', header_bytes, offset=118)[0]
    wfm_info['record_type'] = struct.unpack_from(
        'I', header_bytes, offset=122)[0]
    wfm_info['exp_dim_1_type'] = struct.unpack_from(
        'I', header_bytes, offset=244)[0]
    wfm_info['time_base_1'] = struct.unpack_from(
        'I', header_bytes, offset=768)[0]
    wfm_info['fastframe'] = struct.unpack_from('I', header_bytes, offset=78)[0]
    wfm_info['Frames'] = struct.unpack_from(
        'I', header_bytes, offset=72)[0] + 1
    wfm_info['summary_frame'] = struct.unpack_from(
        'h', header_bytes, offset=154)[0]
    wfm_info['curve_offset'] = struct.unpack_from('i', header_bytes, offset=16)[
        0]  # 838 + ((frames - 1) * 54)
    # scaling factors
    wfm_info['vscale'] = struct.unpack_from('d', header_bytes, offset=168)[0]
    wfm_info['voffset'] = struct.unpack_from('d', header_bytes, offset=176)[0]
    #wfm_info['tstart'] = struct.unpack_from('d', header_bytes, offset=496)[0]
    wfm_info['tstart'] = struct.unpack_from('d', header_bytes, offset=488)[0]
    wfm_info['tscale'] = struct.unpack_from('d', header_bytes, offset=536)[0]
    # trigger detail
    wfm_info['tfrac'] = struct.unpack_from('d', header_bytes, offset=788)[
        0]  # frame index 0
    wfm_info['tdatefrac'] = struct.unpack_from(
        'd', header_bytes, offset=796)[0]  # frame index 0
    wfm_info['tdate'] = struct.unpack_from('I', header_bytes, offset=804)[
        0]  # frame index 0

    # data offsets
    # frames are same size, only first frame offsets are used
    dsize = struct.unpack_from('I', header_bytes, offset=818)[0]
    wfm_info['dsize'] = dsize
    # sample data type detection
    code = struct.unpack_from('i', header_bytes, offset=240)[0]
    wfm_info['code'] = code
    bps = struct.unpack_from('b', header_bytes, offset=15)[
        0]  # bytes-per-sample
    wfm_info['bps'] = bps
    if code == 7 and bps == 1:
        dformat = 'int8'
    elif code == 0 and bps == 2:
        dformat = 'int16'
    elif code == 4 and bps == 4:
        dformat = 'single'
    else:
        raise WfmReadError('data type code or bytes-per-sample not understood')
    wfm_info['dformat'] = dformat
    wfm_info['dlen'] = dsize // bps
    return wfm_info


class ScopeData:

    def __init__(self, path):
        """
        Class for oscillogram data of old Tek Oscilloscopes
        5 Series MSO
        6 Series MSO

        Parameters
        ----------
        path : str
            Full or relative path to oscillogram WFM file.

        Returns
        -------
        None.

        """
        try:
            self.y, tstart, tscale, tfrac = read_wfm(path)
        except Exception as E:
            raise WfmReadError(path, E)
        samples = len(self.y)
        tstop = samples * tscale + tstart
        self.x = np.linspace(tstart, tstop, samples, endpoint=False)
        self.horizInterval = self.x[1] - self.x[0]



from ._numpy_core import *
from ._dcore import P2Q, Q2P
from .histogram import hist2Q, QStatisticsMaker
from .oscillogram import PulsesHistMaker
from .crosstalk import (optimize_pcrosstalk, compensate, distort, 
                        Q2total_pcrosstalk, total_pcrosstalk, ENF)

__version__ = "0.1.3"


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 16 23:10:39 2019

@author: vonGostev
"""
import numpy as np
from scipy.special import binom
from scipy.linalg import pinv
from functools import lru_cache

from ._numpy_core import DPREC


def binomial_t_elem(qe: float, n: int, m: int):
    if m > n:
        return 0
    return qe ** m * (1 - qe) ** (n - m) * binom(n, m)


def binomial_t_matrix(qe: float, N: int, M: int):
    qe = DPREC(qe)
    t_matrix = np.zeros((M, N), dtype=DPREC)
    for n in range(N):
        for m in range(M):
            t_matrix[m, n] = binomial_t_elem(qe, n, m)
    return t_matrix


def binomial_invt_matrix(qe: float, N: int, M: int, n_cells=0):
    qe = DPREC(qe)
    t_matrix = np.zeros((N, M), dtype=DPREC)
    for n in range(N):
        for m in range(M):
            t_matrix[n, m] = binomial_t_elem(1 / qe, m, n)
    return t_matrix


@lru_cache(maxsize=1024)
def subbinom_t_elem(qe: float, n_cells: int, n: int, m: int):
    """
    Calculated from formula (5) via recursive approach from Appendix A

    Sperling, J., Vogel, W., & Agarwal, G. S. (2014).
    Quantum state engineering by click counting.
    Physical Review A, 89(4), 043829.
    for arbitrary quantum efficiency and zero noise

    Idea from
    Sperling, J., Vogel, W., & Agarwal, G. S. (2012).
    True photocounting statistics of multiple on-off detectors.
    Physical Review A, 85(2), 023820.

    """
    if n == 0 and m == 0:
        return 1
    elif n != 0 and m == 0:
        return (1 - qe) ** n
    elif m > n:
        return 0
    return (1 - qe + qe * m / n_cells) * subbinom_t_elem(qe, n_cells, n - 1, m) +\
        qe * (n_cells - m - 1) / n_cells * \
        subbinom_t_elem(qe, n_cells, n - 1, m - 1)


def subbinomial_t_matrix(qe: float, N: int, M: int, n_cells=667):
    qe = DPREC(qe)
    t_matrix = np.zeros((M, N), dtype=DPREC)
    for n in range(N):
        for m in range(M):
            t_matrix[m, n] = subbinom_t_elem(qe, n_cells, n, m)
    return t_matrix


def subbinomial_invt_matrix(qe: float, N: int, M: int, n_cells=667):
    return pinv(subbinomial_t_matrix(qe, N, M, n_cells))
